#include <cmath>

struct Dat{
	double a;
	double b;
	double c;
	double d;
	double x;
	double y;
	double& operator[](unsigned int i)
	{
		switch (i)
		{
			case 0:
				return a;
			case 1:
				return b;
			case 2:
				return c;
			case 3:
				return d;
			case 4:
				return x;
			case 5:
				return y;
		};
		return x;

	}
};

struct XY
{
	double x;
	double y;
};

const double delta=0.025f;
const double diffamount=0.1f;
const double threshold=0.25f;
const int ITERATIONCOUNT=1000;
const double maxjump=100.0f;
const double maxV=100000.0f;

inline Dat inc1(Dat dat)
{
	dat.x+=diffamount;
	return dat;
}

inline Dat dec1(Dat dat)
{
	dat.x-=diffamount;
	return dat;
}

inline Dat inc2(Dat dat)
{
	dat.y+=diffamount;
	return dat;
}

inline Dat dec2(Dat dat)
{
	dat.y-=diffamount;
	return dat;
}

template<int i>
double catastrophe(const Dat& dat);

template<>
double catastrophe<0>(const Dat& dat)
{
	return dat.x*dat.x*dat.x+dat.a*dat.x;
}
template<>
double catastrophe<1>(const Dat& dat)
{
	return dat.x*dat.x*dat.x*dat.x+dat.a*dat.x*dat.x+dat.b*dat.x;
}
template<>
double catastrophe<2>(const Dat& dat)
{
	return dat.x*dat.x*dat.x*dat.x*dat.x+(-dat.a)*dat.x*dat.x*dat.x+dat.b*dat.x*dat.x+dat.c*dat.x;
}
template<>
double catastrophe<3>(const Dat& dat)
{
	double a=dat.a*(amountsign[0]?1:-1);
	double b=dat.b*(amountsign[1]?1:-1);
	double c=dat.c*(amountsign[2]?1:-1);
	double d=dat.d*(amountsign[3]?1:-1);
	return dat.x*dat.x*dat.x*dat.x*dat.x*dat.x+a*dat.x*dat.x*dat.x*dat.x+b*dat.x*dat.x*dat.x+c*dat.x*dat.x+d*dat.x;
}
template<>
double catastrophe<4>(const Dat& dat)
{
	return dat.x*dat.x*dat.x+dat.y*dat.y*dat.y+dat.a*dat.x*dat.y+dat.b*dat.x+dat.c*dat.y;
}
template<>
double catastrophe<5>(const Dat& dat)
{
	double c[3];

	c[amountperm[0]]=dat.a*(amountsign[0]?1:-1);
	c[amountperm[1]]=dat.b*(amountsign[1]?1:-1);
	c[amountperm[2]]=dat.c*(amountsign[2]?1:-1);
	double x;
	double y;
	if (swapxy)
	{
		x=dat.y;
		y=dat.x;
	}
	else
	{
		x=dat.x;
		y=dat.y;
	}
	return x*x*x/3-x*y*y+c[0]*(x*x+y*y)+c[1]*x+c[2]*y;
}
template<>
double catastrophe<6>(const Dat& dat)
{
	double c[4];

	c[amountperm[0]]=dat.a*(amountsign[0]?1:-1);
	c[amountperm[1]]=dat.b*(amountsign[1]?1:-1);
	c[amountperm[2]]=dat.c*(amountsign[2]?1:-1);
	c[amountperm[3]]=dat.d*(amountsign[3]?1:-1);
	double x;
	double y;
	if (swapxy)
	{
		x=dat.y;
		y=dat.x;
	}
	else
	{
		x=dat.x;
		y=dat.y;
	}

	return x*x*y+y*y*y*y+c[0]*x*x+c[1]*y*y+c[2]*x+c[3]*y;
}

template<int i>
XY DCatastrophe(const Dat& dat)
{
	return (XY) {
		diffamount*(catastrophe<i>(inc1(dat))-catastrophe<i>(dec1(dat)))/(2),
		diffamount*(catastrophe<i>(inc2(dat))-catastrophe<i>(dec2(dat)))/(2)
	};
}


//return true if jump in state
template<int n>
MoveInfo move(Dat& dat)
{
	MoveInfo res;



	Dat newdat=dat;
	std::cout<<"STARTCALC\n";
	for (int i=0;i<ITERATIONCOUNT
	                &&abs(newdat.x)<2000
                    && abs(newdat.y)<2000
                    ;i++)
	{
		XY diff = DCatastrophe<n>(newdat);
		newdat.x-=diff.x;
		newdat.y-=diff.y;
	}
	std::cout<<endl;

	//std::cout<<"V "<<catastrophe<n>(newdat)<<endl;
	//std::cout<<"x "<<dat.x<<" -> "<<newdat.x<<endl;
	//std::cout<<"y "<<dat.y<<" -> "<<newdat.y<<endl<<endl;

	double jump=((newdat.x-dat.x)*(newdat.x-dat.x)+(newdat.y-dat.y)*(newdat.y-dat.y));

	double dc = catastrophe<n>(newdat); dc*=dc;

	if (jump>maxjump*maxjump)
	{
		res.movetype=r_fail;
		res.fail.xfail=(newdat.x-dat.x)*(newdat.x-dat.x)>=(maxjump*maxjump/2);
		res.fail.yfail=(newdat.y-dat.y)*(newdat.y-dat.y)>=(maxjump*maxjump/2);

		std::cout<<"FAIL"<<res.fail.xfail<<","<<res.fail.yfail<<endl;
	}
	else if (jump>threshold*threshold)
	{
		res.movetype=r_jump;
		if ((newdat.x-dat.x)*(newdat.x-dat.x)>(newdat.y-dat.y)*(newdat.y-dat.y))
		{//x jump
			res.jump.param=false;
			res.jump.paramdir=newdat.x>dat.x;
		}
		else
		{//y jump
			res.jump.param=true;
			res.jump.paramdir=newdat.y>dat.y;
		}
	}
	else res.movetype=r_none;

	dat=newdat;

	return res;
}
